#include <avr/io.h>
#include <avr/pgmspace.h>
#include <stdlib.h>
#include <stdio.h>
#include <util/delay.h>
#include <stdint.h>
#include <time.h>
#include <string.h>
#include "pcd8544.h"

#include "defs.h"

#define LOST 1
#define PLAY 0
#define TIME_SLOT 1000 // ms (perioada unei intreruperi)
// M, N matricea de joc (nu ecranul complet)
#define M 100
#define N 46

int timp; // timpul total de joc
int punctaj; // punctaj total al jocului
int status;
int m[M][N];
Piesa p;
int start;

void init_mat() {

	int i,j;
	for(i=0; i<N; i++)
		for(j=0; j<M; j++)
			m[i][j] = 0;
}

/* genereaza tipul piesei, seteaza coordonatele */
void init_piesa() {

	 p.tip = (3*timp+2*punctaj)%5;
	 p.rot = 0;
	 switch (p.tip) {
	 
		// coordonatele y = -1 nu vor fi mapate in matrice (si nu vor fi afisate)
		case 0:	// patrat
			p.p1.x = M/2-1; p.p1.y = -1;
			p.p2.x = M/2;   p.p2.y = -1;
			p.p3.x = M/2;   p.p3.y = 0;
			p.p4.x = M/2-1; p.p4.y = 0;
			break;
		
		case 1: // linie
			p.p1.x = M/2; 	p.p1.y = -3;
			p.p2.x = M/2; 	p.p1.y = -2;
			p.p3.x = M/2; 	p.p1.y = -1;
			p.p4.x = M/2; 	p.p1.y = 0;
			break;
			
		case 2: // L
			p.p1.x = M/2-1;	p.p1.y = -2;
			p.p2.x = M/2-1; p.p2.y = -1;
			p.p3.x = M/2-1; p.p3.y = 0;
			p.p4.x = M/2;	p.p4.y = 0;
			break;
			
		case 3: //Z
			p.p1.x = M/2-1;	p.p1.y = -1;
			p.p2.x = M/2;	p.p2.y = -1;
			p.p3.x = M/2;	p.p3.y = 0;
			p.p4.x = M/2+1;	p.p4.y = 0;
			break;
			
		case 4: // T intors
			p.p1.x = M/2;	p.p1.y = -1;
			p.p2.x = M/2-1; p.p2.y = 0;
			p.p3.x = M/2;	p.p3.y = 0;
			p.p4.x = M/2+1;	p.p4.y = 0;
			break;
	 }
}

/* coboara piesa curenta cu o pozitie */
void coboara_piesa () {

	p.p1.y++;
	p.p2.y++;
	p.p3.y++;
	p.p4.y++;

	if(start == 0 && p.p1.y >=0)
		start = 1;
}

/* daca piesa curenta atinge blocurile deja formate si se poate "aseza" */
int asezat () {

	if (m[p.p4.x][p.p4.y+1] == 1 || m[p.p3.x][p.p3.y+1] == 1 ||
		m[p.p2.x][p.p2.y+1] == 1 || m[p.p1.x][p.p1.y+1] == 1) {
			
			// piesa se imprima in matrice
			m[p.p1.x][p.p1.y] = 1;
			m[p.p2.x][p.p2.y] = 1;
			m[p.p3.x][p.p3.y] = 1;
			m[p.p4.x][p.p4.y] = 1;
			return 1;	
	}
	return 0;
}

/* calculeaza punctajul aferent dizolvarii a "n" linii */
int getPoints(int n) {

	if (n == 0)
		return 0;
	return 2*n-1;
}


void check_dizolvari() {

	int i,j;
	int nr_diz = 0;
	for(i=N-1; i>=0; i--) {
	
		int ok = 1;
		for (j=0; j<M; j++)
			if (m[i][j] == 0) {
				ok = 0;
				break;
			}
		if (ok == 1) { // linia i se poate dizolva
			int k;
			for(k = i; k>0; k--)
				for(j=0; j<M; j++)
					m[k][j] = m[k-1][j];
			for(j=0; j<M; j++)
				m[0][j] = 0;
			nr_diz++;
		}	
	}
	punctaj += getPoints(nr_diz);
}


/* iesire din joc; scrie rezultat statistic */
void game_over() {

	status = LOST;
	LcdClear ();
	LcdGotoXYFont (20,40);
	
	int i;
	char * str = strdup ("Game over");
	for (i=0; i<strlen(str); i++) {
	
		byte ch = (byte) str[i];
		LcdChr (FONT_1X, byte ch);
	}
	LcdGotoXYFont (25,30);
	
	int new_punctaj = 0, len = 0; // inversat
	while (punctaj != 0) {
	
		new_punctaj *= 10;
		new_punctaj += punctaj % 10;
		punctaj /= 10;
		len++;
	}
	
	// scrie punctaj de la stg -> dreapta
	for (i=0; i<len; i++) {
	
		byte ch = (byte) (new_punctaj%10);
		new_punctaj /= 10;
		LcdChr (FONT_1X, byte ch);
	}
}


/* cand una din subpiese are coordonata y < 0 (nu mai poate patrunde pe ecranul jocului) 
* SAU
* piesa depaseste limitele laterale */
int out_borders(Piesa ps) {

	if (ps.p1.y < 0 || ps.p2.y < 0 && start == 1)
		return 1;
	if (ps.p1.x < 0 || ps.p2.x < 0 || ps.p3.x >= M || ps.p4.x >= M)
		return 1;
	return 0;
}


/* translateaza piesa */
Piesa translate(int mode) {

	Piesa aux = p;
	
	switch (mode) {
	
		case 1:
			// stanga
			aux.p1.x--;
			aux.p2.x--;
			aux.p3.x--;
			aux.p4.x--;
			break;
			
		case 2:
			// dreapta
			aux.p1.x++;
			aux.p2.x++;
			aux.p3.x++;
			aux.p4.x++;
			break;
	}
	return aux;
}


/* daca piesa se suprapune peste ceva in lateral */
int intersect(Piesa ps) {

	if(m[ps.p1.x][ps.p1.y] || m[ps.p2.x][ps.p2.y] || m[ps.p3.x][ps.p3.y] || m[ps.p4.x][ps.p4.y])
		return 1;
	return 0;
}


/* rotatie la 90 grd. trigo. */
Piesa rotate (Piesa ps) {

	int aux1, aux2;
	switch(ps.tip) {
	
		case 0:
			// patratul ramane la fel
			break;

		case 1:
			// linia
			if (ps.rot == 0) {
				ps.rot = 1;
				ps.p1.x = ps.p2.x - 1; 
				ps.p1.y = ps.p2.y;
				ps.p3.x = ps.p2.x + 1;
				ps.p3.y = ps.p2.y;
				ps.p4.x = ps.p2.x + 2;
				ps.p4.y = ps.p2.y;
			}
			else {
				ps.rot = 0;
				ps.p1.x = ps.p2.x;
				ps.p1.y = ps.p2.y - 1;
				ps.p3.x = ps.p2.x + 1;
				ps.p3.y = ps.p2.y;
				ps.p4.x = ps.p2.x + 2;
				ps.p4.y = ps.p2.y;
			}
			break;
			
		case 2:
			// L
			switch (ps.rot) {
			
				case 0:
					ps.rot++;
					ps.p1.x = ps.p3.x -1;
					ps.p1.y = ps.p3.y;
					ps.p2 = ps.p3;
					ps.p3 = ps.p4;
					ps.p4.x = ps.p3.x - 1;
					ps.p4.y = ps.p3.y;
					break;
					
				case 1:
					ps.rot++;
					ps.p4 = ps.p3;
					ps.p3.x = ps.p4.x;
					ps.p3.y = ps.p4.y - 1;
					ps.p2.x = ps.p3.x;
					ps.p2.y = ps.p3.y - 1;
					ps.p1.x = ps.p2.x - 1;
					ps.p1.y = ps.p1.y;
					break;
					
				case 2:
					ps.rot++;
					ps.p3 = ps.p1;
					ps.p4 = ps.p2;
					ps.p2.x = ps.p3.x - 1;
					ps.p2.y = ps.p3.y;
					ps.p1.x = ps.p2.x;
					ps.p1.y = ps.p2.y + 1;
					break;
					
				case 3:
					ps.rot = 0;
					struct patratel aux = ps.p1;
					ps.p2 = ps.p1;
					ps.p1 = aux;
					ps.p3.x = ps.p1.x;
					ps.p3.y = ps.p1.y + 1;
					ps.p4.x = ps.p3.x + 1;
					ps.p4.y = ps.p4.y;
					break;
			}
			break;
			
		case 3:
			// Z
			if (ps.rot == 0) {
				ps.rot++;
				ps.p2 = ps.p3;
				ps.p3 = ps.p4;
				ps.p1 = ps.p4;
				ps.p1.y--;
				ps.p4 = ps.p3;
				ps.p4.y++;
			}
			else {
				ps.rot--;
				ps.p4 = ps.p3;
				ps.p3 = ps.p2;
				ps.p2 = ps.p3;
				ps.p2.y--;
				ps.p1 = ps.p2;
				ps.p1.x--;
			}
			break;
			
		case 4:
			// T
			switch (ps.rot) {
			
				case 0:
					ps.rot++;
					struct patratel aux;
					aux = ps.p1;
					ps.p1 = ps.p2;
					ps.p2 = aux;
					ps.p4 = ps.p3;
					ps.p4.y++;
					break;
					
				case 1:
					ps.rot++;
					ps.p2 = ps.p3;
					ps.p3 = ps.p2;
					ps.p3.x++;
					break;
					
				case 2:
					ps.rot++;
					ps.p1 = ps.p2;
					ps.p1.y--;
					break;
					
				case 3:
					ps.rot=0;
					ps.p4 = ps.p3;
					ps.p3 = ps.p2;
					ps.p2 = ps.p3;
					ps.p2.x--;
					ps.p1 = ps.p2;
					ps.p1.x++;
					ps.p1.y--;
					break;
			}
			break;
	}
	return ps;
}

/* deseneaza corespondentul unui element din matrice */
void print_patratel (struct patratel p) {

	// un patratel are suprafata 3 x 3 pixeli
	byte x = (byte) (3*p.x);	// 3 = offsetul acumulat
	byte y = (byte) (3*p.y);
	
	LcdPixel (x, y, PIXEL_ON);
	LcdPixel (x, y+1, PIXEL_ON);
	LcdPixel (x, y+2, PIXEL_ON);
	
	LcdPixel (x+1, y, PIXEL_ON);
	LcdPixel (x+1, y+1, PIXEL_ON);
	LcdPixel (x+1, y+2, PIXEL_ON);
	
	LcdPixel (x+2, y, PIXEL_ON);
	LcdPixel (x+2, y+1, PIXEL_ON);
	LcdPixel (x+2, y+2, PIXEL_ON);
}

/* deseneaza matricea curenta */
void afisare () {

	int i,j;
	for (i=0; i<N; i++)
		for (j=0; j<M; j++)
			if (m[i][j] == 1) {
				struct patratel pt;
				pt.x = i;
				pt.y = j;
				print_patratel(pt);
			}
	LcdUpdate();
}


/* ce se intampla in fiecare cuanta de timp */
int interrupt () {

	_delay_ms(TIME_SLOT);
	timp++;
	int cod = button_code();
	if (cod == 0) {

		coboara_piesa();
		if (asezat() == 0) ;
		else {
			// s-a asezat; exista dizolvari?
			check_dizolvari();
			// iese din limita superioara (nu mai poate intra)
			if(out_borders(p))
				return 1;
			// piesa s-a asezat cu succes, nu depaseste limita ecranului;
			// intra o noua piesa
			init_piesa();
		}
	}
	else {
	
		// a fost apasat un buton
		switch (cod) {
			
			case 1: case 2:		// stg, dreapta
				Piesa aux = translate(cod);
				if (out_borders(aux) || intersect(aux) == 1) ; // nu se accepta mutarea
				else
					p = aux;
				break;
				
			case 3:				// asezare
				do {
					coboara_piesa();
				} while (asezat() == 0);
				break;
				
			case 4:				// rotatie
				Piesa aux = rotate(p);
				if (out_borders(aux) || intersect (aux) == 0)
					p = aux;
				break;
				
			case -1:			// terminare joc
				game_over();
				return 0;
				break;
			case 3: 
				while (asezat()==0)
					coboara_piesa ();
				// s-a asezat; exista dizolvari?
				check_dizolvari();
				// iese din limita superioara (nu mai poate intra)
				if(out_borders(p))
					return 1;
				// piesa s-a asezat cu succes, nu depaseste limita ecranului;
				// intra o noua piesa
				init_piesa();
				break;
		}
	}
	
	LcdClear();
	afisare();
	
	return 0;
}


int main () {

	status = PLAY; // still playing...
	init_mat();
	init_piesa();
	start = 0;	// devine 1 cand toate subpiesele au y >=0
	LcdInit();
	LcdClear();
	
	while (status != LOST) {

		int rc = interrupt();
		if (rc == 1 && status != LOST) // s-a terminat
			game_over();
	}

	return 0;
}
